/*
 * Abugraph.java
 *
 * Copyright (c) 2007 Ignacio Labrador Pavon
 * 28814 Daganzo, Madrid.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL COPYRIGHT HOLDERS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package net.sourceforge.abugraph;

import javax.swing.JApplet;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StreamCorruptedException;
import java.util.Date;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import javax.swing.filechooser.FileSystemView;

/**
 * Abugraph GUI implementation.
 * 
 * @author Ignacio Labrador Pavon
 * @since  June 2007
 */
public class Abugraph extends JApplet implements ActionListener, ItemListener
{
  //~Constants ----------------------------------------------------------------
  
  //General purpouse
  private final static String VERSION          = "7.10.14";
  private final static String NAME_VERSION     = "AbuGraph! " + VERSION;
  private final static int    INI_PANEL_WIDTH  = 1100;
  private final static int    INI_PANEL_HEIGHT = 800;
  private final static String DASH_REGEXP      = "\\-";
  private final static String DEF_EDGE_LABEL   = "-";
  private final static String EMPTY_LABEL      = "         ";
  private final static String FILES_SUFFIX     = ".aua";
  private final static String CRLF             = "\r\n";
  
  //Button labels
  private final static String SCRAMBLE    = "Scramble";
  private final static String SHAKE       = "Shake";
  private final static String ZOOM_IN     = "Zoom+";
  private final static String ZOOM_OUT    = "Zoom-";  
  private final static String CLEAR       = "Clear";
  private final static String EXIT        = "Exit";
  private final static String SAVE        = "Save";
  private final static String LOAD        = "Load";
  private final static String LABELS      = "Labels";
  private final static String PAUSE       = "Pause";
  private final static String HELP        = "Help";
  private final static Font   BUTTON_FONT = new Font("default",Font.BOLD,11);
  
  //Remote commands
  private final static char   R_UPDATE = 'U';
  private final static char   R_REMOVE = 'R';
  private final static char   R_HELP   = 'H';
  private final static char   R_QUIT   = 'Q';
  private final static char   R_NEW    = 'N';
  private final static char   R_CLEAR  = 'C';
  
  //Messages 
  private final static String WELCOME        = "Welcome to the AbuGraph! "+VERSION+" server. "; 
  private final static String TYPE           = "Type '"; 
  private final static String FOR_HELP       = "' for help";
  private final static String BYE            = "Bye!";
  private final static String ERROR          = "Error! ";
  private final static String INVALID_CMD    = "Unexpected command"; 
  private final static String INVALID_SINTAX = "Invalid sintax";
  private final static String USAGE          = "Usage:";
  private final static String CMD            = "java -jar AbuGraph.jar -port=[port]";
  private final static String PORT_RANGE     = "The port must be an integer from 1025 to 65535";
  private final static String NO_GUI         = "Unable to start-up the GUI.";
  
  //The frame
  private final static JFrame frame = new JFrame(NAME_VERSION);
  
  //Look and feel strings (just two of them are used at the moment)
  private final static String MOTIF   = "com.sun.java.swing.plaf.motif.MotifLookAndFeel";
  private final static String METAL   = "javax.swing.plaf.metal.MetalLookAndFeel";
  private final static String WINDOWS = "com.sun.java.swing.plaf.windows.WindowsLookAndFeel";
  private final static String GTK     = "com.sun.java.swing.plaf.gtk.GTKLookAndFeel";
  
  //Command line arguments
  private final static String LISTENING_PORT = "-port";
  
  //~Variables ----------------------------------------------------------------
  private AbugraphPanel          panel           = null;
  private JPanel                 controlPanel    = null;
  private JButton                scramble        = new JButton(SCRAMBLE);
  private JButton                shake           = new JButton(SHAKE);
  private JButton                zoomIn          = new JButton(ZOOM_IN);
  private JButton                zoomOut         = new JButton(ZOOM_OUT);
  private JButton                clear           = new JButton(CLEAR);
  private JButton                save            = new JButton(SAVE);
  private JButton                exit            = new JButton(EXIT);
  private JButton                load            = new JButton(LOAD);
  private JButton                help            = new JButton(HELP);
  private JCheckBox              labels          = new JCheckBox(LABELS);
  private JCheckBox              pause           = new JCheckBox(PAUSE);
  private ConfirmationDialog     dialog          = null;
  private String                 dialogCaller    = null;
  private static int             server_port     = -1;
  private static JFileChooser    fileChooser     = null;

  //~Methods ------------------------------------------------------------------
  
  /**
   * The main method (althoug the class was devised as an applet, it can
   * be executed as an application as well). 
   */
  public static void main(String[] args)
  {
    // Deal with incoming arguments
    if(args.length == 0)
    {
      System.out.print(USAGE+CRLF);
      System.out.print(CMD+CRLF);
      System.exit(-1);
    }
    
    //Set applet stub and get command line parameters
    final JApplet applet = new Abugraph();
    try
    {
      applet.setStub(new AbugraphAppletStub(args, applet));
      server_port = Integer.parseInt(applet.getParameter(LISTENING_PORT));
      if(server_port <= 1024 || server_port > 65535) throw new Exception();
    }
    catch(Exception e)
    {
      System.out.print(CRLF);
      System.out.print(USAGE+CRLF);
      System.out.print(CMD+CRLF);      
      System.out.print(CRLF);
      System.out.print(PORT_RANGE+CRLF);
      System.out.print(CRLF);
      System.exit(-1);
    }
    
    //Get the file system view to manage OS depending variables
    FileSystemView fileSystemView = FileSystemView.getFileSystemView();
    fileChooser = new JFileChooser( fileSystemView.getHomeDirectory() );
      
    //
    // Start up the GUI
    // 
   
    // Define look and feel
    try
    {
      String laf = (File.separatorChar == '/') ? METAL : WINDOWS;
      javax.swing.UIManager.setLookAndFeel(laf);
    } 
    catch (Exception ex)
    {
      System.out.print(ERROR+NO_GUI+CRLF);
      ex.printStackTrace(); //TODO: remove
      System.exit(-1);
    } 
    
    // Window listener to close the window
    frame.addWindowListener(new WindowAdapter()
    {
      public void windowClosing(WindowEvent evt)
      {
        applet.stop();
        applet.destroy();
        System.exit(0);
      }
    });    
    
    //Add regular components to the window and start-up the GUI
    frame.setSize(INI_PANEL_WIDTH,INI_PANEL_HEIGHT);
    frame.add(BorderLayout.CENTER, applet);
    frame.getContentPane().add(applet);  
    applet.init();
    frame.setVisible(true);
    applet.start();  
  }

  /**
   * Applet init method
   */
  public void init()
  { 
    setLayout(new BorderLayout());
    panel = new AbugraphPanel(this);
    panel.setBackground(Color.white);
    dialog = new ConfirmationDialog(frame,this); 
        
    add(BorderLayout.CENTER, panel);
    controlPanel = new JPanel();
    add(BorderLayout.NORTH, controlPanel);
    
    load.setFont(BUTTON_FONT);
    controlPanel.add(load);
    load.addActionListener(this);
    
    save.setFont(BUTTON_FONT);
    controlPanel.add(save);
    save.addActionListener(this);

    clear.setFont(BUTTON_FONT);
    controlPanel.add(clear);
    clear.addActionListener(this);

    help.setFont(BUTTON_FONT);
    controlPanel.add(help);
    help.addActionListener(this);
    
    exit.setFont(BUTTON_FONT);
    controlPanel.add(exit);
    exit.addActionListener(this); 
    
    controlPanel.add(new JLabel(EMPTY_LABEL));
    
    scramble.setFont(BUTTON_FONT);
    controlPanel.add(scramble);
    scramble.addActionListener(this);
    scramble.setForeground(Color.BLUE);
    
    shake.setFont(BUTTON_FONT);
    controlPanel.add(shake);
    shake.addActionListener(this);
    shake.setForeground(Color.BLUE);

    zoomIn.setFont(BUTTON_FONT);
    controlPanel.add(zoomIn);
    zoomIn.addActionListener(this);
    zoomIn.setForeground(Color.BLUE);    
    
    zoomOut.setFont(BUTTON_FONT);
    controlPanel.add(zoomOut);
    zoomOut.addActionListener(this);
    zoomOut.setForeground(Color.BLUE);    
    
    controlPanel.add(new JLabel(EMPTY_LABEL));
    
    labels.setFont(BUTTON_FONT);
    controlPanel.add(labels);
    labels.addItemListener(this);
    
    pause.setFont(BUTTON_FONT);
    controlPanel.add(pause);
    pause.addItemListener(this);    
  }

  /**
   * Applet destroy method
   */
  public void destroy()
  {
    remove(panel);
    remove(controlPanel);
  }

  /**
   * Returns a reference to the panel
   */
  public AbugraphPanel getPanel()
  {
    return this.panel;
  }
  
  /**
   * Method to get the client request code. This method is used because some
   * telnet clients send a null prior to the actual message
   *
   * @para ar_receivedCmd String with the received command as returned by the
   *                      'readLine()' method
   *
   * @return If the received command starts with 0x00, a substring without that
   *         initial character; otherwhise the received command parameter itself
   */
  private String getCommand(String ar_receivedCmd)
  {
    StringBuilder sb = new StringBuilder();
    
    int start;
    if(ar_receivedCmd!=null && (byte)ar_receivedCmd.charAt(0)==0x00) start = 1;
    else start = 0;
    
    //Remove non-printable characteres
    for(int i=start; i<ar_receivedCmd.length(); i++)
    {
      char c = ar_receivedCmd.charAt(i);
      if(c > 0x1F && c < 0x7F) sb.append(c);
      else if(c == 0x08) sb.deleteCharAt(sb.length()-1); //backspace to DEL                                     
    }
    
    return sb.toString();
  }
  
  /**
   * Applet start method. Used to start the GUI and the graph server
   */
  public void start()
  {
    Server server = null;
    String clientRequest = null;
    BufferedReader reader = null;
    PrintWriter writer = null;

    panel.start();
    
    boolean connected = false;
    while (true)
    {      
      if(!connected)
      {
        server = new Server(server_port);
        reader = new BufferedReader(new InputStreamReader(server.in));
        writer = new PrintWriter(new OutputStreamWriter(server.out), true);
        writer.println(WELCOME + new Date() + CRLF);
        writer.println(TYPE + R_HELP + FOR_HELP + CRLF);
        connected = true;
      }
      
      try
      {
        while ((clientRequest = getCommand(reader.readLine()) ) != null)
        {
          if(clientRequest.length() == 0) continue;
          
          //TRACE (just for debugging....)
          //for(int i=0; i<clientRequest.length(); i++)
          //  System.out.println("Received : ["+(byte)clientRequest.charAt(i)+']');

          char requestCode = clientRequest.charAt(0);
          if(requestCode == R_REMOVE)
          {
            String[] params = clientRequest.split(DASH_REGEXP);
            panel.removeEdge(params[1],params[2]);
            continue;
          }
          else if(requestCode == R_UPDATE)
          {
            String[] params = clientRequest.split(DASH_REGEXP);
            if(params.length == 3)
            {
              panel.updateNode(params[1],params[2]);
              continue;
            }
            else if(params.length == 4)
            {
              panel.updateEdge(params[1],params[2],params[3]);
              continue;
            }
            else
            {
              writer.println(ERROR + INVALID_CMD + CRLF);
              continue;
            }            
          }         
          else if(requestCode == R_NEW)
          {
            //Example: fresa-naranja-fruta
            String[] params = clientRequest.split(DASH_REGEXP);
            if(params.length == 3) 
              panel.addEdge(params[1],params[2],DEF_EDGE_LABEL);
            else
              panel.addEdge(params[1],params[2],params[3]);
          }
          else if(requestCode == R_CLEAR)
          {
            panel.clear();
          }
          else if(requestCode == R_QUIT)
          {
            server.close();
            connected = false;    
            break;
          }          
          else if(requestCode == R_HELP)
          {
            writer.println( this.getHelpMessage(true) );
            continue;
          }            
          else
          {
            writer.println(ERROR + INVALID_CMD + CRLF);
            continue;
          }
        }
        
        String close_msg = "Ups! The remote client closed the connection...";
        MessageDialog msg = new MessageDialog(frame,close_msg);
        msg.setVisible(true); 

        server.close();
        connected = false;        
      }
      catch (Exception e)
      {
        writer.println(ERROR + INVALID_SINTAX + CRLF);
        continue;         
      }
    }      
  }
  
  /**
   * Return the dialog caller
   */
  public boolean wasCalledToExit()
  {
    return this.dialogCaller.equals(EXIT);
  }

  /**
   * Return the dialog caller
   */
  public boolean wasCalledToClear()
  {
    return this.dialogCaller.equals(CLEAR);
  }
  
  /**
   * Method to generate the help message
   *
   * @param ar_usage Boolean to decide if the word "usage" will appears at
   *                 the begining of the help message
   */
  private String getHelpMessage(boolean ar_usage)
  {
    String msg = "";
    if(ar_usage) msg += CRLF + "Usage:" + CRLF;
    msg += 
      CRLF+" **To add a new edge connecting two nodes write \"N-orig-dest\"."+CRLF+
      "   Example: \"N-one-two\" will create the node \"one\" with a link"+CRLF+
      "             towards the node \"two\"." +
      CRLF+CRLF+
      " **To add a new labeled edge connecting two nodes write"+CRLF+
      "   \"N-orig-dest-label\"." + CRLF +
      "   Example: \"N-one-two-1\" will create the node \"one\" linked with"+CRLF+
      "             the node \"two\" being \"1\" the label for the new edge."+
      CRLF+CRLF+
      " **To remove an existing edge use \"R-origin-destination\""+CRLF+
      "   Example: \"R-one-two\" will remove the edge between the nodes"+CRLF+            
      "            \"one\" and \"two\"." +
      CRLF+CRLF+
      " **To update an existing node use \"U-CurrentLabel-NewLabel\"."+CRLF+
      "   Example: \"U-one-first\" will modify the node label \"one\" with"+CRLF+ 
      "             the new label \"first\"." +
      CRLF+CRLF+
      " **To update an existing edge label use \"U-Orig-Dest-NewLabel\"."+CRLF+
      "   Example: \"U-one-two-I\" will modify the edge connecting nodes"+CRLF+
      "            \"one\" and \"two\", being \"I\" the new label value." +            
      CRLF+CRLF+
      " **To clear the current graph use \"C\"." +            
      CRLF+CRLF+
            " **To quit use \"Q\""+CRLF;
    return msg;
  }
  
  /**
   * Applet stop method
   */
  public void stop()
  {
    panel.stop();
  }

  /**
   * Event handling method
   */
  public void actionPerformed(ActionEvent e)
  {
    Object src = e.getSource();
    if (src.equals(scramble))
    {
      Dimension d = getSize();
      for( Map.Entry me : panel.getNodes().entrySet() )
      {
        Node n = (Node)me.getValue();
        if ( !n.isFixed() )
        {
          n.setX(10 + (d.width - 20) * Math.random());
          n.setY(10 + (d.height - 20) * Math.random());
        }
      }
      return;
    }
    
    else if (src.equals(shake))
    {
      for( Map.Entry me : panel.getNodes().entrySet() )
      {
        Node n = (Node)me.getValue();
        if ( !n.isFixed() )
        {
          double rand = 80 * Math.random() - 40;
          n.addToX(rand);
          n.addToY(rand);
        }
      }
    }

    else if (src.equals(zoomIn))
    {
      panel.zoomIn();
    }

    else if (src.equals(zoomOut))
    {
      panel.zoomOut();
    }    
    
    else if (src.equals(load))
    { 
      int openDialog = fileChooser.showOpenDialog(Abugraph.this);
      if(openDialog == JFileChooser.APPROVE_OPTION)
      {
        String filename = fileChooser.getCurrentDirectory().toString() + 
                          File.separatorChar +
                          fileChooser.getSelectedFile().getName();
        this.load(filename);
        panel.defineZoomLevel();
      } 
    }
    
    else if (src.equals(save))
    {
      int saveDialog = fileChooser.showSaveDialog(Abugraph.this);
      if(saveDialog == JFileChooser.APPROVE_OPTION)
      {
        String filename = fileChooser.getCurrentDirectory().toString() +
                          File.separatorChar +
                          fileChooser.getSelectedFile().getName();
   
        this.save(filename); 
      }      
    }
    
    else if (src.equals(clear))
    {
      dialogCaller = CLEAR;
      dialog.setVisible(true);
    }
    
    else if (src.equals(exit))
    {
      dialogCaller = EXIT;
      dialog.setVisible(true);
    }
    
    else if (src.equals(help))
    {
      String str = "Just open a text console and type \"telnet localhost " + 
                   server_port + "\"."+CRLF+
                   "Then you can type the following commands:"+CRLF;
      str += this.getHelpMessage(false);
      MessageDialog msg = new MessageDialog(frame,str);
      msg.setVisible(true);
    }    
  }

  /**
   * Method to save the current graph state in a file.
   * 
   * @param ar_file Filename (specifying the full path)
   */
  private void save(String ar_file)
  {    
    if( !ar_file.toLowerCase().endsWith(FILES_SUFFIX) ) ar_file += FILES_SUFFIX;

    String errorMsg = null;
    try
    {
      Object[] graphState = panel.getGraphState();
      if(graphState == null) throw new NullPointerException();
      FileOutputStream fos = new FileOutputStream(ar_file);
      GZIPOutputStream gzos = new GZIPOutputStream(fos);    
      ObjectOutputStream out = new ObjectOutputStream(gzos);
      out.writeObject(graphState);     
      out.flush();            
      out.close(); 
      return;
    }
    catch (FileNotFoundException e)
    {
      errorMsg = ERROR + "Unable to access to that path.";     
    }    
    catch(SecurityException e)
    {
      errorMsg = ERROR + "Permission denied.";
    }
    catch(IllegalArgumentException e)
    {
      errorMsg = ERROR + "You are trying to generate an empty file.";
    }
    catch(NullPointerException e)
    {
      errorMsg = ERROR + "You are trying to save an empty graph.";
    }
    catch (Exception e)
    { 
      errorMsg = ERROR + "Unexpected error trying to save file.";
    }

    MessageDialog msg = new MessageDialog(frame,errorMsg);
    msg.setVisible(true);      
  }  
  
  /**
   * Load a graph from a file.
   *
   * @param ar_file Filename string (specified with the full path)
   */
  private void load(String ar_file)
  {
    if(!ar_file.toLowerCase().endsWith(FILES_SUFFIX) )
    {
      String message = "Error. Suffix \"" + FILES_SUFFIX + "\" expected.";
      MessageDialog msg = new MessageDialog(frame,message);
      msg.setVisible(true);
      return;
    }
    
    String errorMsg = null;
    try
    {
      FileInputStream fis = new FileInputStream(ar_file);
      GZIPInputStream gzis = new GZIPInputStream(fis);  
      ObjectInputStream in = new ObjectInputStream(gzis); 

      Object[] graph_file_info = (Object[])in.readObject();
      panel.setGraphState(graph_file_info);
      in.close(); 
      repaint();
      return;
    }
    catch (FileNotFoundException e)
    {
      errorMsg = ERROR + "Unable to access to that file.";     
    }    
    catch(SecurityException e)
    {
      errorMsg = ERROR + "Permission denied.";
    }
    catch(StreamCorruptedException e)
    {
      errorMsg = ERROR + "File corrupted.";
    }
    catch(NullPointerException e)
    {
      errorMsg = ERROR + "Empty file.";
    }    
    catch(InvalidClassException e)
    {
      errorMsg = ERROR + "Unexpected file format.";
    }
    catch(ClassNotFoundException e)
    {
      errorMsg = ERROR + "Unrecognized file format.";
    }
    catch(ClassCastException e)
    {
      errorMsg = ERROR + "Invalid file format.";
    }
    catch (Exception e)
    { 
      errorMsg = ERROR + "Unexpected error trying to read file.";
    }

    MessageDialog msg = new MessageDialog(frame,errorMsg);
    msg.setVisible(true); 
  }
  
  /**
   * Applet itemStateChanged method
   */
  public void itemStateChanged(ItemEvent e)
  {
    Object src = e.getSource();
    boolean on = e.getStateChange() == ItemEvent.SELECTED;
    if (src == labels) panel.setLabels(on);
    else if (src == pause) panel.setPause(on);
  }

  /**
   * Method to provide applet information
   */
  public String getAppletInfo()
  {
    return "Title: AbuGraph! " + NAME_VERSION + CRLF + CRLF +
           "The server is listeing for connections on port " +
           Integer.toString(server_port);
  }

  /**
   * Method to provide parameter information
   */
  public String[][] getParameterInfo()
  {
    String[][] info =
    {
      { "Connection port", "int", Integer.toString(server_port) }
    };
    return info;
  } 
}
