/*
 * AbugraphPanel.java
 *
 * Copyright (c) 2004 Sun Microsystems, Inc. 
 * All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * -Redistribution of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 * 
 * -Redistribution in binary form must reproduce the above copyright notice, 
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of contributors may 
 * be used to endorse or promote products derived from this software without 
 * specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any kind. ALL 
 * EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING
 * ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. SUN MIDROSYSTEMS, INC. ("SUN")
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE
 * AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST 
 * REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, 
 * INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND REGARDLESS OF THE THEORY 
 * OF LIABILITY, ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, 
 * EVEN IF SUN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * 
 * You acknowledge that this software is not designed, licensed or intended
 * for use in the design, construction, operation or maintenance of any
 * nuclear facility.
 */

package net.sourceforge.abugraph;

import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Panel;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * AbuGraph panel class
 */
class AbugraphPanel extends Panel 
implements Runnable, MouseListener, MouseMotionListener
{
  //~Constants ----------------------------------------------------------------
  private static final int    DEF_ZOOM_LEVEL   = 50;
  private static final int    ZOOM_STEP        = 10;
  
  //~Variables ----------------------------------------------------------------
  private Abugraph              graph;
  private Map<String, Integer>  node_labels;   
  private Map<Integer, Node>    nodes;         
  private Set<Edge>             edges;
  private Thread                thread;
  private boolean               showEdgeLabels;
  private boolean               pause;
  private int                   numMouseButtonsDown;
  private Node                  selected;
  private boolean               pickfixed;
  private Image                 offscreen;
  private Dimension             offscreensize;
  private Graphics              offgraphics;
  private Set<Integer>          releasedNodeIndexes;
  private int                   zoomLevel;
  
  //~Methods ------------------------------------------------------------------
  
  /**
   * Abugraph constructor (to be replaced when menu works)
   */
  public AbugraphPanel(Abugraph ar_graph)
  {
    this.graph = ar_graph;
    this.node_labels = new HashMap<String, Integer>();
    this.edges = new HashSet<Edge>();
    this.releasedNodeIndexes = new HashSet<Integer>();
    this.nodes = new HashMap<Integer, Node>();
    this.zoomLevel = DEF_ZOOM_LEVEL;
    addMouseListener(this);
  }
  
  /**
   * Method to clear the panel
   */
  public void clear()
  {
    this.edges.clear();
    this.nodes.clear();
    this.node_labels.clear();
    this.releasedNodeIndexes.clear();
  }
  
  /**
   * This method is called after loading a file. It is used to recover the
   * zoom level used by that file graph.
   */
  public void defineZoomLevel()
  {
    if( !this.edges.isEmpty() )
    {
      int sum_lengths = 0;
      for(Edge e: this.edges) sum_lengths += e.getLength();
      this.zoomLevel = sum_lengths/this.edges.size();
    }
    else this.zoomLevel = DEF_ZOOM_LEVEL;
  }

  /**
   * Method to zoomIn
   */
  public void zoomIn()
  {
    this.zoomLevel += ZOOM_STEP;
    for(Edge e: this.edges) e.setLength(this.zoomLevel);
  }
  
  /**
   * Method to zoomOut
   */
  public void zoomOut()
  {
    this.zoomLevel -= ZOOM_STEP; 
    if(this.zoomLevel < DEF_ZOOM_LEVEL) this.zoomLevel = DEF_ZOOM_LEVEL;
    for(Edge e: this.edges) e.setLength(this.zoomLevel);
  }
  
  /**
   * Method to define de 'showEdgeLabels' attribute
   */
  public void setLabels(boolean arg)
  {
    this.showEdgeLabels = arg;
  }
  
  /**
   * Method to define de 'pause' attribute
   */
  public void setPause(boolean arg)
  {
    this.pause = arg;
  }
  
  /**
   * Method to get the nodes collection
   */
  public Map<Integer, Node> getNodes()
  {
    return this.nodes;
  }
  
  /**
   * Method to get a new node index. It tries to re-use a discarded index, but
   * if it is not possible, it generates a new one
   */
  private int getNextNodeIndex()
  {
    int rv;
    if(this.releasedNodeIndexes.isEmpty())
    {
      rv = this.nodes.size();
    }
    else
    {
      rv = this.releasedNodeIndexes.iterator().next();
      this.releasedNodeIndexes.remove(rv);
    }
    return rv;
  }
  
  /**
   * Method to add a new node
   * @param ar_label Node label
   * 
   * @return The node index
   */
  public int addNode(String ar_label)
  {
    int nodeIdx = this.getNextNodeIndex();
    Node n = new Node(ar_label);
    if(nodeIdx == 0) n.setFixed(true);
    n.setX(10 + 380 * Math.random());
    n.setY(10 + 380 * Math.random());
    this.nodes.put(nodeIdx,n);
    this.node_labels.put(ar_label,nodeIdx);
    return nodeIdx;
  }

  /**
   * Method to update the edges label
   *
   * @param ar_orig     The edge originator node
   * @param ar_dest     The edge destination node
   * @param ar_newLabel The new label
   */
  public void updateEdge(String ar_orig, String ar_dest, String ar_newLabel)
  {
    Integer origIndex = this.node_labels.get(ar_orig);
    if(origIndex == null) return;
    
    Integer destIndex = this.node_labels.get(ar_dest);
    if(destIndex == null) return;
    
    Edge e = new Edge(origIndex, destIndex, zoomLevel);
    if(this.edges.remove(e) == true)
    {
      e.setLabel(ar_newLabel);
      this.edges.add(e);
    }
  }
  
  /**
   * Method to update the node label
   *
   * @param ar_label    Current label (the one to be replaced)
   * @param ar_newLabel The new label
   */
  public void updateNode(String ar_label, String ar_newLabel)
  {    
    Integer nodeToUpdate = this.node_labels.get(ar_label);
    if(nodeToUpdate != null) 
    {     
      Node node2update = nodes.get(nodeToUpdate);
      node2update.setLabel(ar_newLabel);
      
      Integer nodeExist = this.node_labels.get(ar_newLabel);
      if(nodeExist != null) 
      {
        Node existingNode = this.nodes.get(nodeExist);
        
        Set<Integer> existingFromNodes = existingNode.getFromNodes();
        for(Integer from2remove : existingFromNodes)
        {
          Edge e2remove = new Edge(from2remove,nodeExist,zoomLevel);
          this.edges.remove(e2remove);
          Edge e2add = new Edge(from2remove,nodeToUpdate,zoomLevel);
          this.edges.add(e2add);
          node2update.addFrom(from2remove);
        }
        
        Set<Integer> existingToNodes = existingNode.getToNodes();        
        for(Integer to2remove : existingToNodes)
        {
          Edge e2remove = new Edge(nodeExist,to2remove,zoomLevel);
          this.edges.remove(e2remove);
          Edge e2add = new Edge(nodeToUpdate,to2remove,zoomLevel);
          this.edges.add(e2add); 
          node2update.addTo(to2remove);
        }        
        
        this.removeNode(nodeExist);
      }
      
      this.node_labels.remove(ar_label);
      this.node_labels.put(ar_newLabel,nodeToUpdate);
    }
  }
  
  /**
   * Method to remove an edge
   * 
   * @param ar_from Originating node
   * @param ar_to   Destination node
   */
  public void removeEdge(String ar_from,String ar_to)
  { 
    Integer from = this.node_labels.get(ar_from);
    if(from == null) return;

    Integer to = this.node_labels.get(ar_to);
    if(to == null) return;    
    
    Edge e = new Edge(from,to,zoomLevel);

    if( this.edges.remove(e) == true)
    {
      Node nodeFrom = nodes.get(e.getFrom());
      if(nodeFrom != null)
      {
        nodeFrom.decNeighbours();
        if( !nodeFrom.hasNeighbours() ) removeNode(e.getFrom());
        else nodeFrom.removeTo(to);
      }
      
      Node nodeTo = nodes.get(e.getTo());
      if(nodeTo != null)
      {
        nodeTo.decNeighbours();
        if(!nodeTo.hasNeighbours()) removeNode(e.getTo());
        else nodeTo.removeFrom(from);
      }
    }
  }
  
  /**
   * Method to remove a node from the graph
   * 
   * @param ar_nodeIdx Node index 
   */
  public void removeNode(int ar_nodeIdx)
  {
    Node node2del = this.nodes.get(ar_nodeIdx);
    this.node_labels.remove(node2del.getLabel());
    this.nodes.remove(new Integer(ar_nodeIdx));
    this.releasedNodeIndexes.add(ar_nodeIdx);
  }

  /**
   * Method to add a new edge 
   * 
   * @param ar_from  Originating node
   * @param ar_to    Destination node
   * @param ar_label The edge label
   * 
   * @return The just created edge
   */
  public Edge addEdge(String ar_from, String ar_to, String ar_label)
  {
    //System.out.println("+++ Adding edge from ["+from+"] to ["+to+']')
    
    Integer from = this.node_labels.get(ar_from);
    if(from == null) from = this.addNode(ar_from);
    
    Integer to = this.node_labels.get(ar_to);
    if(to == null) to = this.addNode(ar_to);
    
    Edge e = new Edge(from,to,zoomLevel,ar_label);

    if( edges.add(e) == true )
    {
      Node fromNode = this.nodes.get(from);
      fromNode.incNeighbours();
  
      Node toNode = this.nodes.get(to);  
      toNode.incNeighbours();
      
      fromNode.addTo(to);
      toNode.addFrom(from);
    }

    return e;
  }

  /**
   * Thread 'run' method. Basically checks the pause flag and calls the
   * 'relax()' method periodically.
   */
  public void run()
  {
    Thread me = Thread.currentThread();
    while (thread == me)
    {
      while(this.pause == true)  
      {
        try { Thread.sleep(1000); }
        catch (InterruptedException e) {break;}        
      }

      relax();

      try { Thread.sleep(100); }
      catch (InterruptedException e) {break;}
    }
  }

  /**
   * "Relaxing" method. Used to periodically update the graph.
   */
  private synchronized void relax()
  {
    for(Edge e : this.edges)
    {
      Node nodeTo = this.nodes.get(e.getTo());
      Node nodeFrom = this.nodes.get(e.getFrom());
      
      if(nodeTo == null || nodeFrom == null) continue;
      
      double vx = nodeTo.getX() - nodeFrom.getX();
      double vy = nodeTo.getY() - nodeFrom.getY();
      double len = Math.sqrt(vx*vx + vy*vy);
      len = (len == 0) ? .0001 : len;
      double f = (e.getLength() - len) / (len * 3);
      double dx = f * vx;
      double dy = f * vy;
      nodeTo.setDx(dx + nodeTo.getDx());
      nodeTo.setDy(dy + nodeTo.getDy());
      nodeFrom.setDx(nodeFrom.getDx() - dx);
      nodeFrom.setDy(nodeFrom.getDy() - dy);
    }
    
    for( Map.Entry me : nodes.entrySet() )
    { 
      Node n1 = (Node)me.getValue();
      double dx = 0;
      double dy = 0;
      for( Map.Entry mf : nodes.entrySet() )
      {
        Node n2 = (Node)mf.getValue();
        if (n1.equals(n2)) continue;

        double vx = n1.getX() - n2.getX();
        double vy = n1.getY() - n2.getY();
        double len = vx * vx + vy * vy;
        if (len == 0)
        {
          dx += Math.random();
          dy += Math.random();
        }
        else if (len < 100 * 100)
        {
          dx += vx / len;
          dy += vy / len;
        }
      }
      double dlen = dx * dx + dy * dy;
      if (dlen > 0)
      {
        dlen = Math.sqrt(dlen) / 2;
        n1.setDx(n1.getDx()+(dx/dlen));
        n1.setDy(n1.getDy()+(dy/dlen));
      }
    }
    
    Dimension d = getSize();
    for( Map.Entry me : nodes.entrySet() )
    {
      Node n = (Node)me.getValue();
      if ( !n.isFixed() )
      {
        n.addToX(Math.max(-5, Math.min(5, n.getDx())));
        n.addToY(Math.max(-5, Math.min(5, n.getDy())));
      }
      
      if (n.getX() < 0)
      {
        n.setX(0);
      }
      else if (n.getX() > d.width)
      {
        n.setX(d.width);
      }
      
      if (n.getY() < 0)
      {
        n.setY(0);
      }
      else if (n.getY() > d.height)
      {
        n.setY(d.height);
      }
      
      n.setDx(n.getDx() / 2.0);
      n.setDy(n.getDy() / 2.0);
    }
    repaint();
  }

  /**
   * Thread start method
   */
  public void start()
  {
    thread = new Thread(this);
    thread.start();
  }

  /**
   * Thread stop method
   */
  public void stop()
  {
    thread = null;
  }        
  
  /**
   * Update method
   * 
   * @param ar_g Graphics object reference
   */
  public synchronized void update(Graphics ar_g)
  {
    Dimension d = getSize();
    if ((offscreen == null) || (d.width != offscreensize.width)
        || (d.height != offscreensize.height))
    {
      offscreen = createImage(d.width, d.height);
      offscreensize = d;
      if (offgraphics != null)
      {
        offgraphics.dispose();
      }
      offgraphics = offscreen.getGraphics();
      //offgraphics.setFont(NODE_LABEL_FONT);
    }
    offgraphics.setColor(getBackground());
    offgraphics.fillRect(0, 0, d.width, d.height);

    //Draw all edges and nodes
    for(Edge e : edges) e.draw(offgraphics, this.nodes, this.showEdgeLabels);
    FontMetrics fm = offgraphics.getFontMetrics();
    for( Map.Entry me : nodes.entrySet() )
    {
      Node n = (Node)me.getValue();
      n.draw(offgraphics, fm, selected);
    }
    ar_g.drawImage(offscreen, 0, 0, null);
  }

  /**
   * Method to get the current graph state, that is, the data structures
   * which allows re-build the graph after deserialization from a file.
   * 
   * 
   * @return Array of Objects containig:
   *         <ul>
   *         <li> The nodes map as first element
   *         <li> The edges set as second element
   *         <li> The releases node indexes set as third element
   *         <li> The node showEdgeLabels map as fourth element 
   *         </ul>
   */
  public Object[] getGraphState()
  {
    Object[] rv = null;
    if( !this.edges.isEmpty() )
    {
      rv = new Object[4];
      rv[0] = this.nodes;
      rv[1] = this.edges;
      rv[2] = this.releasedNodeIndexes;
      rv[3] = this.node_labels;
    }
    
    return rv;
  }
  
  /**
   * Method to set the graph state (this is the complementary method of
   * 'getGraphState()'
   * 
   * 
   * @param Array of Objects containig:
   *         <ul>
   *         <li> The nodes map as first element
   *         <li> The edges set as second element
   *         <li> The releases node indexes set as third element
   *         <li> The node showEdgeLabels map as fourth element 
   *         </ul>
   */
  public void setGraphState(Object[] ar_state)
  {
    this.nodes = (Map<Integer,Node>)ar_state[0];
    this.edges = (Set<Edge>)ar_state[1];
    this.releasedNodeIndexes = (Set<Integer>)ar_state[2];
    this.node_labels = (Map<String,Integer>)ar_state[3];
  }

  //
  // Event handling methods 
  //
  
  /**
   * Invoked when the mouse button has been clicked (pressed
   * and released) on a component.
   */
  public void mouseClicked(MouseEvent e) {}

  /**
   * Invoked when a mouse button has been pressed on a component.
   */
  public void mousePressed(MouseEvent e)
  {
    numMouseButtonsDown++;
    addMouseMotionListener(this);
    double bestdist = Double.MAX_VALUE;
    int x = e.getX();
    int y = e.getY();
    for( Map.Entry me : nodes.entrySet() )
    {
      Node n = (Node)me.getValue();
      double dist = (n.getX()-x) * (n.getX()-x) + (n.getY()-y) * (n.getY()-y);
      if (dist < bestdist)
      {
        selected = n;
        bestdist = dist;
      }
    }
    if(selected != null)
    {
      pickfixed = selected.isFixed();
      selected.setFixed(true);
      selected.setX(x);
      selected.setY(y);
    }
    repaint();
    e.consume();
  }

  /**
   * Invoked when a mouse button has been released on a component.
   */
  public void mouseReleased(MouseEvent e)
  {
    numMouseButtonsDown--;
    removeMouseMotionListener(this);
    if(selected != null)
    {
      selected.setFixed(pickfixed);
      selected.setX( e.getX() );
      selected.setY( e.getY() );
      if (numMouseButtonsDown == 0) selected = null;
    }
    repaint();
    e.consume();
  }

  /**
   * Invoked when the mouse enters a component.
   */
  public void mouseEntered(MouseEvent e) {}

  /**
   * Invoked when the mouse exits a component.
   */
  public void mouseExited(MouseEvent e) {}

  /**
   * Invoked when a mouse button is pressed on a component and then 
   * dragged.  <code>MOUSE_DRAGGED</code> events will continue to be 
   * delivered to the component where the drag originated until the 
   * mouse button is released (regardless of whether the mouse position 
   * is within the bounds of the component).
   * <p> 
   * Due to platform-dependent Drag&Drop implementations, 
   * <code>MOUSE_DRAGGED</code> events may not be delivered during a native 
   * Drag&Drop operation.  
   */
  public void mouseDragged(MouseEvent e)
  {
    selected.setX( e.getX() );
    selected.setY( e.getY() );
    repaint();
    e.consume();
  }

  /**
   * Invoked when the mouse cursor has been moved onto a component
   * but no buttons have been pushed.
   */
  public void mouseMoved(MouseEvent e) {}  
}
