/*
 * Edge.java
 *
 * Copyright (c) 2007 Ignacio Labrador Pavon
 * 28814 Daganzo, Madrid.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL COPYRIGHT HOLDERS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package net.sourceforge.abugraph;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Polygon;
import java.io.Serializable;
import java.util.Map;

/**
 * Edge class. 
 *
 * @author ilp
 * @since  June-2007
 */
public class Edge implements Serializable
{
  //~Constants ----------------------------------------------------------------
  private static final long   serialVersionUID = -3721211867649111L;
  private static final Color  EDGE_LABEL_COLOR = Color.blue;
  private static final Color  ARC_COLOR_1      = Color.black;
  private static final Color  ARC_COLOR_2      = Color.lightGray;
  private static final Color  ARC_COLOR_3      = Color.red;
  private static final Font   EDGE_LABEL_FONT  = new Font("default",Font.ITALIC,10);
  private static final int    ARROW_SIZE       = 8;
  private static final double ARROW_HOOK       = 0.8;
  
  //~Variables ----------------------------------------------------------------
  private Integer from   = null;
  private Integer to     = null;
  private String  label  = null;
  private int     hash   = 0;
  private int     length = 0;
  
  //~Methods ------------------------------------------------------------------
  
  /**
   * Constructor (without label)
   *
   * @param from Originating node
   * @param to   Destination node
   */
  public Edge(int from, int to, int length)
  {
    this.from = from;
    this.to = to;
    this.label = null;
    this.length = length;
  }
  
  /**
   * Constructor 
   *
   * @param from  Originating node
   * @param to    Destination node
   * @param label Edge label
   */
  public Edge(int from, int to, int length, String label)
  {
    this.from = from;
    this.to = to;
    this.label = label;
    this.length = length;
  }
  
  /**
   * Overrides the Object.hashcode method. Just the members 'from' and 'to' are
   * used to calculate the hash code.
   */
  public int hashCode()
  {
    if(hash == 0)
    {
      StringBuilder sb = new StringBuilder();
      sb.append(from);
      sb.append('-');
      sb.append(to);
      hash = sb.toString().hashCode();
    }
    return hash;
  }
  
  /**
   * Overrides the Object.equals method
   */
  public boolean equals(Object other)
  {
    if(other instanceof Edge) return (this.hash == other.hashCode());
    else throw new ClassCastException(); 
  }
  
  /**
   * Method to return the edge associated value
   */
  public String getLabel()
  {
    return this.label;
  }
  
  /**
   * Method to define the edge associated value
   */
  public void setLabel(String lbl)
  {
    this.label = lbl;
  }
  
  /**
   * Method to return the 'from'
   */
  public Integer getFrom()
  {
    return this.from;
  }
 
  /**
   * Method to return the 'to'
   */
  public Integer getTo()
  {
    return this.to;
  }  

  /**
   * Method to return the 'to'
   */
  public int getLength()
  {
    return this.length;
  }
  
  /**
   * Method to return the 'to'
   */
  public void setLength(int ar_length)
  {
    this.length = ar_length;
  }   
  
  /**
   * Method to draw an edge consisting on a line with an arrow, and a label.
   *
   * @param g          Graphics object where the edge is painted
   * @param nodes      Nodes map (relates node with node index)
   * @param showLabels Flag to request edge labels drawing
   */
  public void draw(Graphics g, Map<Integer, Node> nodes, boolean showLabels)
  {   
    Node nodeFrom = nodes.get(this.from);
    Node nodeTo = nodes.get(this.to);

    if(nodeFrom == null || nodeTo == null) return;

    int x0 = (int) nodeFrom.getX();
    int y0 = (int) nodeFrom.getY();
    int x1 = (int) nodeTo.getX();
    int y1 = (int) nodeTo.getY();
    int len = (int)Math.abs(Math.sqrt((x0-x1)*(x0-x1)+(y0-y1)*(y0-y1))-this.length);

    Color c = (len < 10) ? ARC_COLOR_1 
                         : len < 20 ? ARC_COLOR_2 
                                    : ARC_COLOR_3;
    g.setColor(c);    

    int dx = x1 - x0;
    int dy = y1 - y0;
    double d = Math.sqrt((dx*dx)+(dy*dy));    
    
    if(d != 0) 
    {
      //The line
      g.drawLine(x0, y0, x1, y1);    
      
      //The edge labels (if requested)
      int midX0 = (x1+x0)/2;
      int midY0 = (y1+y0)/2;
      if (showLabels && this.label != null)
      {
        Font prevFont = g.getFont();
        Color prevColor = g.getColor();
        g.setFont(EDGE_LABEL_FONT);
        g.setColor(EDGE_LABEL_COLOR);
        int font_height = g.getFontMetrics().getHeight();
        int font_width = g.getFontMetrics().charWidth('x');
        int x_label = ((x1+midX0)/2) + font_width;
        int y_label = (y1+midY0)/2;
        if(y1 > y0 && x1 < x0) y_label += font_height;
        else if(y1 < y0 && x1 > x0) y_label += font_height;
        g.drawString(this.label,x_label,y_label);
        g.setFont(prevFont);
        g.setColor(prevColor);
      }      

      // The arrow
      Polygon p = new Polygon();
      dx = (int)(dx / d * ARROW_SIZE);
      dy = (int)(dy / d * ARROW_SIZE);
      x1 = (x1+midX0) / 2;
      y1 = (y1+midY0) / 2;
      p.addPoint(x1,y1);
      p.addPoint(x1-(int)(dx-(dy/2.0)), y1-(int)((dx/2.0)+dy));
      p.addPoint(x1-(int)((dx*ARROW_HOOK)), y1-(int)((dy*ARROW_HOOK)));
      p.addPoint(x1-(int)((dy/2.0)+dx), y1-(int)(dy-(dx/2.0)));
      g.fillPolygon(p);
      g.drawPolygon(p);
    }
  }         
  
  /**
   * Returns a string containig the edge information
   */
  public String toString()
  {
    StringBuilder sb = new StringBuilder();
    sb.append(from);
    sb.append(" -> ");
    sb.append(to);
    sb.append(" [");
    sb.append(this.label);
    sb.append(']');
    return sb.toString();
  }
}
