/*
 * Node.java
 *
 * Copyright (c) 2007 Ignacio Labrador Pavon
 * 28814 Daganzo, Madrid.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL COPYRIGHT HOLDERS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package net.sourceforge.abugraph;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Set;

/**
 * Node class
 * 
 * @author ilp
 * @since July-2007
 */
public class Node implements Comparable, Serializable
{ 
  //~Constants ----------------------------------------------------------------
  private static final long  serialVersionUID  = 1429861986214L;
  private static final Color FIXED_NODE_COLOR  = Color.orange;
  private static final Color DEF_NODE_COLOR    = Color.yellow;
  private static final Color SELECT_NODE_COLOR = Color.pink;
  private static final Font  NODE_LABEL_FONT  = new Font("default",Font.BOLD,10);
  
  //~Variables ----------------------------------------------------------------
  private int          hash = 0;
  private double       x;
  private double       y;
  private double       dx;
  private double       dy;
  private boolean      fixed;
  private String       label;
  private int          neighbours;
  private Set<Integer> from;
  private Set<Integer> to;
  private Color        userColor;

  //~Methods ------------------------------------------------------------------

  /**
   * Void constructor
   */
  public Node()
  {
    this.neighbours = 0;
    this.fixed = false;
    this.from = new HashSet<Integer>();
    this.to = new HashSet<Integer>();
  }

  /**
   * Constructor with a label as argument
   */
  public Node(String ar_label)
  {
    this.neighbours = 0;
    this.fixed = false;
    this.label = ar_label;
    this.from = new HashSet<Integer>();
    this.to = new HashSet<Integer>();
  }  
  
  /**
   * Method to add to the 'x' member the specified value
   */
  public void addToX(double arg)
  {
    this.x += arg;
  }   

  /**
   * Method to add to the 'y' member the specified value
   */
  public void addToY(double arg)
  {
    this.y += arg;
  }  
  
  /**
   * Method to define the 'x' member
   */
  public void setX(double arg)
  {
    this.x = arg;
  }  

  /**
   * Method to define the 'y' member
   */
  public void setY(double arg)
  {
    this.y = arg;
  }   

  /**
   * Method to get the 'x' member
   */
  public double getX()
  {
    return this.x;
  }  

  /**
   * Method to get the 'y' member
   */
  public double getY()
  {
    return this.y;
  }  
  
  /**
   * Method to get the 'label' member
   */
  public String getLabel()
  {
    return this.label;
  }  
  
  /**
   * Method to define the 'label' member
   */
  public void setLabel(String arg)
  {
    this.label = arg;
  }  
  
  /**
   * Method to fix a position for the node
   */
  public void setFixed(boolean arg)
  {
    this.fixed = arg;
  }

  /**
   * Method to check if a node has a fixed position
   */
  public boolean isFixed()
  {
    return this.fixed;
  }

  /**
   * Method to increase the number of node neighbours
   */
  public void incNeighbours()
  {
    if(this.neighbours < Integer.MAX_VALUE) this.neighbours++;
  }  

  /**
   * Method to check if a node has neighbour nodes
   */
  public boolean hasNeighbours()
  {
    return (this.neighbours > 0);
  }  
  
  /**
   * Method to decrease the number of node neighbours
   */
  public void decNeighbours()
  {
    if(this.neighbours > 0) this.neighbours--;
  }   
  
  /**
   * Method to get the 'dx' member
   */
  public double getDx()
  {
    return this.dx;
  }

  /**
   * Method to get the 'dy' member
   */
  public double getDy()
  {
    return this.dy;
  }  
  
  /**
   * Method to define the 'dx' member
   */
  public void setDx(double arg)
  {
    this.dx = arg;
  }  

  /**
   * Method to define the 'dy' member
   */
  public void setDy(double arg)
  {
    this.dy = arg;
  }  
  
  /**
   * Overrides the Object.hashcode() method
   */
  public int hashCode()
  {
    if(this.hash == 0)
    {
      StringBuffer sb = new StringBuffer();
      sb.append(label);;
      this.hash = sb.toString().hashCode();
    }
    return hash;
  }
  
  /**
   * Overrides the Object.equals() method
   */
  public boolean equals(Object other)
  {
    if(other instanceof Node) return (this.hash == other.hashCode());
    else throw new ClassCastException(); 
  }

  /** 
   * @see java.lang.Comparable#compareTo(java.lang.Object)
   */
  public int compareTo(Object other)
  {
    if(other instanceof Node)
    {
      Node o = (Node)other;
      return this.label.compareTo(o.label); 
    }
    else throw new ClassCastException(); 
  }
  
  /**
   * Method to add a new pointed node.
   *
   * @return 'true' if the argumet is actually a new pointed node, 'false'
   *         otherwise.
   */
  public boolean addTo(int ar_to)
  {
    return this.to.add(ar_to);
  }

  /**
   * Method to add a new pointing node.
   *
   * @return 'true' if the argumet is actually a new pointed node, 'false'
   *         otherwise.
   */
  public boolean addFrom(int ar_from)
  {
    return this.from.add(ar_from);
  }
  
  /**
   * Method to remove a pointing node.
   */
  public void removeFrom(int ar_from)
  {
    this.from.remove(ar_from);
  }

  /**
   * Method to remove a pointed node.
   */
  public void removeTo(int ar_to)
  {
    this.to.remove(ar_to);
  }  
  
  /**
   * Method to retreive the from nodes set
   */
  public Set<Integer> getFromNodes()
  {
    return this.from;
  }

  /**
   * Method to retreive the to nodes set
   */
  public Set<Integer> getToNodes()
  {
    return this.to;
  }  
  
  /**
   * Method to paint the node
   * 
   * @param ar_g        Graphics object where the node is painted
   * @param ar_fm       Font metrics
   * @param ar_selected Selected node (if this node is the ar_selected one, it is
   *                    painted in a different way)
   */
  public void draw(Graphics ar_g, FontMetrics ar_fm, Node ar_selected)
  {
    int x = (int)this.x;
    int y = (int)this.y;
    int w = ar_fm.stringWidth(this.label) + 10;
    int h = ar_fm.getHeight() + 4;

    ar_g.setFont(NODE_LABEL_FONT);
    
    Color c = (userColor !=null) ? userColor :
              (this == ar_selected) ? SELECT_NODE_COLOR
                                    : this.isFixed() ? FIXED_NODE_COLOR 
                                                     : DEF_NODE_COLOR;
    ar_g.setColor(c);
    ar_g.fillOval(x-w/2, y-h/2, w, h);
    ar_g.setColor(Color.black);
    ar_g.drawOval(x - w / 2, y - h / 2, w - 1, h - 1);
    ar_g.drawString(this.label,x-(w-10)/2,(y-(h-4)/2)+ar_fm.getAscent());
  }  
  
  /**
   * Returns a string with the node data
   */
  public String toString()
  {
    StringBuilder sb = new StringBuilder();
    sb.append(label);
    sb.append(" -Neigbours:");
    sb.append(this.neighbours);
    return sb.toString();
  }

  void setColor(Color color) {
    this.userColor = color;
  }
}
